from typing import Any

from pytest_matchers.matchers import IsInstance, Matcher
from pytest_matchers.matchers.matcher_factory import matcher
from pytest_matchers.utils.matcher_utils import as_matcher
from pytest_matchers.utils.repr_utils import concat_reprs, non_capitalized


@matcher
class StrictDict(Matcher):
    def __init__(
        self,
        matching: dict,
        *,
        extra_condition: bool = None,
        when_true: dict = None,
        when_false: dict = None,
    ):
        super().__init__()
        self._is_instance_matcher = IsInstance(dict)
        self._matching = matching
        if (when_true or when_false) and extra_condition is None:
            raise ValueError("extra_condition must be set when using when_true or when_false")
        self._extra_condition = extra_condition
        self._when_true = when_true or {}
        self._when_false = when_false or {}

    def matches(self, value: Any) -> bool:
        return self._is_instance_matcher == value and self._expected_dictionary() == value

    def _expected_dictionary(self) -> dict:
        expected_value = self._matching
        if self._extra_condition is not None:
            if self._extra_condition:
                expected_value |= self._when_true
            else:
                expected_value |= self._when_false
        return expected_value

    def __repr__(self) -> str:
        if self._expected_dictionary() == {}:
            return "To be an empty dictionary"
        return concat_reprs("To be a dictionary", self.expectations_repr())

    def expectations_repr(self):
        matching_reprs = [
            f"expecting {repr(key)} {as_matcher(value).concatenated_repr()}"
            for key, value in self._expected_dictionary().items()
        ]
        return non_capitalized(concat_reprs("only", *matching_reprs))



from typing import Any

import pytest
from pytest_matchers.matchers.base import Matcher
from pytest_matchers.utils.matcher_detector import MatcherDetector


def _matches_operation(operation, actual: Any, expected: Matcher) -> bool:
    if operation == "==":
        return expected == actual
    if operation == "!=":
        return expected != actual
    return False  # pragma: no cover


@pytest.hookimpl()
def pytest_assertrepr_compare(config, op: str, left: Any, right: Any):
    del config
    if MatcherDetector(right).uses_matchers() and _matches_operation(op, left, right):
        return [
            f"{left!r} {op} {right!r}",
            "",
            "WARNING! "
            "Comparison failed because the left object redefines the equality operator.",
            "Consider using the 'assert_match' or 'assert_not_match' functions instead.",
            "assert_match(actual, expected)",
        ]
    return None



from typing import Any, Callable

from pytest_matchers.matchers.base import Matcher
from pytest_matchers.utils.matcher_utils import as_matcher
from pytest_matchers.utils.repr_utils import concat_reprs


class AfterApplying(Matcher):
    def __init__(self, function: Callable, expected_value: Any | Matcher):
        super().__init__()
        self._function = function
        self._expected_value = as_matcher(expected_value)

    def matches(self, value: Any) -> bool:
        try:
            return self._function(value) == self._expected_value
        except Exception:  # pylint: disable=broad-except
            return False

    def __repr__(self):
        return concat_reprs(
            f"After applying the {self._function.__name__} function, the result is expected",
            self._expected_value,
        )



from typing import Any

from pytest_matchers.matchers.base import Matcher
from pytest_matchers.matchers.matcher_factory import matcher
from pytest_matchers.utils.matcher_utils import as_matcher


@matcher
class Not(Matcher):
    def __init__(self, value: Matcher | Any):
        super().__init__()
        self._value = as_matcher(value)

    def __new__(cls, value: Matcher | Any):
        if isinstance(value, Not):
            return value._value
        return super().__new__(cls)

    def __invert__(self) -> Matcher:
        return self._value

    def matches(self, value: Any) -> bool:
        return not self._value == value

    def __repr__(self) -> str:
        return f"To not be {self._value.concatenated_repr()}"



from .asserts.asserts import assert_match, assert_not_match
from .main import (
    after_applying_is,
    anything,
    between,
    case,
    contains,
    different_value,
    has_attribute,
    if_false,
    if_true,
    is_datetime,
    is_datetime_string,
    is_dict,
    is_exception,
    is_float,
    is_instance,
    is_int,
    is_iso_8601_date,
    is_iso_8601_datetime,
    is_iso_8601_time,
    is_json,
    is_list,
    is_number,
    is_strict_dict,
    is_string,
    is_timestamp,
    is_uuid,
    not_empty_string,
    one_of,
    same_value,
)

try:
    from .pydantic.main import is_pydantic, is_pydantic_v1
except ImportError:  # pragma: no cover
    pass



from builtins import Exception

from pytest_matchers.matchers.base import Matcher
from pytest_matchers.utils.matcher_utils import as_matcher_or_none, matches_or_none
from pytest_matchers.utils.repr_utils import concat_matcher_repr


class ExceptionMatcher(Matcher):
    def __init__(
        self,
        exception_type: type[Exception] = Exception,
        message: str | Matcher | None = None,
        match_subclass: bool = False,
    ):
        super().__init__()
        self._exception_type = exception_type
        self._message_matcher = as_matcher_or_none(message)
        self._match_subclass = match_subclass

    def matches(self, value: Exception) -> bool:
        return self._matches_type(value) and self._matches_message(value)

    def __repr__(self) -> str:
        representation = "To be an exception"
        if self._match_subclass:
            representation += " instance of"
        else:
            representation += " of type"
        representation += f" {self._exception_type.__name__}"
        if self._message_matcher:
            representation += f" with message expected {concat_matcher_repr(self._message_matcher)}"
        return representation

    def _matches_type(self, value: Exception) -> bool:
        if self._match_subclass:
            return isinstance(value, self._exception_type)
        return type(value) == self._exception_type  # pylint: disable=unidiomatic-typecheck

    def _matches_message(self, value: Exception) -> bool:
        return matches_or_none(self._message_matcher, str(value))



from .after_applying import AfterApplying
from .and_matcher import And
from .anything import Anything
from .base import Matcher
from .between import Between
from .case import Case
from .contains import Contains
from .datetime import Datetime
from .datetime_string import DatetimeString
from .dict import Dict
from .different_value import DifferentValue
from .ends_with import EndsWith
from .eq import Eq
from .exception import ExceptionMatcher
from .has_attribute import HasAttribute
from .if_matcher import If
from .is_instance import IsInstance
from .json import JSON
from .length import Length
from .list import List
from .matcher_factory import MatcherFactory
from .not_matcher import Not
from .number import Number
from .or_matcher import Or
from .same_value import SameValue
from .starts_with import StartsWith
from .strict_dict import StrictDict
from .string import String
from .timestamp import Timestamp
from .uuid import UUID



from typing import Any, Set

from pytest_matchers.asserts.comparers.base import BaseComparer


class SetComparer(BaseComparer):
    def compare(self) -> bool:
        same_length = len(self._actual) == len(self._expected)
        if not same_length and self._fail_fast:
            return False

        return self._items_compare() and same_length

    def _items_compare(self) -> bool:
        expected = self._expected.copy()
        compared_items = (self._in(actual_value, expected) for actual_value in self._actual)
        if not self._fail_fast:
            compared_items = list(compared_items)
        return all(compared_items) and not expected

    def _in(self, actual_value: Any, expected: Set[Any]) -> bool:
        for expected_value in expected:
            if self._base_compare(actual_value, expected_value):
                expected.discard(expected_value)
                return True
        return False



from typing import Type

from pydantic import BaseModel
from pydantic.v1 import BaseModel as BaseModelV1
from pytest_matchers.pydantic.matchers import PydanticModel


def is_pydantic(
    model_class: Type[BaseModel | BaseModelV1] | None = None,
    *,
    strict: bool = None,
    exempt_defaults: bool = True,
    attributes: dict = None,
    **kwargs,
):
    model_class = model_class or BaseModel
    if strict is None:
        strict = bool(attributes) or bool(kwargs)
    return PydanticModel(
        model_class,
        strict=strict,
        exempt_defaults=exempt_defaults,
        attributes=attributes,
        **kwargs,
    )


def is_pydantic_v1(
    model_class: Type[BaseModelV1] | None = None,
    *,
    strict: bool = None,
    exempt_defaults: bool = True,
    attributes: dict = None,
    **kwargs,
):
    return is_pydantic(
        model_class or BaseModelV1,
        strict=strict,
        exempt_defaults=exempt_defaults,
        attributes=attributes,
        **kwargs,
    )



from datetime import datetime
from typing import Any

from pytest_matchers.matchers.base import Matcher
from pytest_matchers.matchers.between import between_matcher
from pytest_matchers.matchers.matcher_factory import matcher
from pytest_matchers.matchers.number import Number
from pytest_matchers.utils.matcher_utils import matches_or_none
from pytest_matchers.utils.repr_utils import concat_reprs


def _as_timestamp(value: float | datetime) -> float:
    if isinstance(value, datetime):
        return value.timestamp()
    return value


@matcher
class Timestamp(Matcher):
    def __init__(self, *, min_value: float | datetime = None, max_value: float | datetime = None):
        super().__init__()
        self._instance_matcher = Number()
        self._limit_matcher = between_matcher(
            _as_timestamp(min_value),
            _as_timestamp(max_value),
            None,
            None,
            None,
        )

    def matches(self, value: Any) -> bool:
        return self._instance_matcher == value and matches_or_none(self._limit_matcher, value)

    def __repr__(self) -> str:
        return concat_reprs("To be a timestamp", self._limit_matcher)



from typing import IO

try:
    from _pytest._io.pprint import PrettyPrinter
except ImportError:  # pragma: no cover
    from pprint import PrettyPrinter

from pytest_matchers.matchers.base import Matcher


def _print_matcher(
    _printer: PrettyPrinter,
    matcher_instance: Matcher,
    stream: IO[str],
    _indent: int,
    _allowance: int,
    _context: set[int],
    _level: int,
):
    stream.write(matcher_instance.next_compared_value_repr())


class MatcherFactory:
    matchers = {}

    @classmethod
    def register(cls, matcher_class: Matcher):
        PrettyPrinter._dispatch[matcher_class.__repr__] = (  # pylint: disable=protected-access
            _print_matcher
        )
        cls.matchers[matcher_class.__name__] = matcher_class

    @classmethod
    def get(cls, name):
        return cls.matchers[name]


def matcher(matcher_class):
    MatcherFactory.register(matcher_class)
    return matcher_class



from typing import Any, Callable

from pytest_matchers.matchers.base import Matcher
from pytest_matchers.matchers.eq import Eq


def as_matcher(value: Matcher | Any) -> Matcher:
    if isinstance(value, Matcher):
        return value
    return Eq(value)


def as_matcher_or_none(value: Matcher | Any | None) -> Matcher | None:
    if value is None:
        return None
    return as_matcher(value)


def matches_or_none(matcher: Matcher | None, value: Any) -> bool:
    return matcher is None or matcher == value


def partial_matches_or_none(matcher: Matcher | None) -> Callable[[Any], bool]:
    def _partial_matches_or_none(value: Any) -> bool:
        return matches_or_none(matcher, value)

    return _partial_matches_or_none



from collections.abc import Sized
from typing import Any

from pytest_matchers.matchers.base import Matcher
from pytest_matchers.matchers.matcher_factory import matcher
from pytest_matchers.utils.matcher_detector import MatcherDetector
from pytest_matchers.utils.repr_utils import as_matcher_repr


def _ends_with_matcher(prefix: Matcher, value: Any):
    if prefix == value:  # With some luck the matcher matches with the whole value
        return True
    try:
        for index in range(1, len(value)):
            if prefix == value[-index:]:
                return True
    except TypeError:
        pass
    return False


@matcher
class EndsWith(Matcher):
    def __init__(self, suffix: Sized | str | Matcher):
        super().__init__()
        self._suffix = suffix

    def matches(self, value: Any) -> bool:
        if MatcherDetector(self._suffix).uses_matchers():
            return _ends_with_matcher(self._suffix, value)

        if isinstance(value, str):
            try:
                return value.endswith(self._suffix)
            except TypeError:
                return False
        if isinstance(value, list):
            suffix_len = len(self._suffix)
            return value[-suffix_len:] == self._suffix

        return False

    def __repr__(self) -> str:
        return f"To end with {repr(self._suffix)}"

    def concatenated_repr(self) -> str:
        return f"with ending expected {as_matcher_repr(self._suffix)}"


def ends_with_matcher(ends_with: str | None) -> EndsWith | None:
    if ends_with is None:
        return None
    return EndsWith(ends_with)



from typing import Any, Type

from pydantic import BaseModel
from pydantic.fields import FieldInfo
from pydantic.v1 import BaseModel as BaseModelV1
from pydantic.v1.fields import ModelField
from pytest_matchers.matchers import HasAttribute, Matcher
from pytest_matchers.matchers.is_instance import is_instance_matcher
from pytest_matchers.matchers.matcher_factory import matcher
from pytest_matchers.utils.repr_utils import concat_reprs


def _model_fields(model_class: Type[BaseModel | BaseModelV1]):
    if issubclass(model_class, BaseModelV1):
        return model_class.__fields__
    return model_class.model_fields


def _is_required(field: FieldInfo | ModelField) -> bool:
    if isinstance(field, ModelField) or not hasattr(field, "is_required"):
        return field.required
    return field.is_required()


def _default_value(field: FieldInfo | ModelField) -> Any:
    if field.default_factory is None:
        return field.default
    return field.default_factory()


def _version(model_class: Type) -> str | None:
    is_pydantic = hasattr(model_class, "__pydantic_complete__") or hasattr(
        model_class, "__fields__"
    )
    if is_pydantic:
        if hasattr(model_class, "__pydantic_complete__"):
            return "v2"
        return "v1"
    return None


@matcher
class PydanticModel(Matcher):
    def __init__(
        self,
        model_class: Type[BaseModel | BaseModelV1],
        *,
        strict: bool = True,
        exempt_defaults: bool = True,
        attributes: dict = None,
        **kwargs,
    ):
        super().__init__()
        self._model_class = model_class
        self._is_instance_matcher = is_instance_matcher(model_class)
        self._strict = strict
        self._exempt_defaults = exempt_defaults
        self._attributes = kwargs
        self._attributes.update(attributes or {})
        self._assert_matching_all_attributes()

    def _assert_matching_all_attributes(self):
        if not self._strict:
            return
        fields = _model_fields(self._model_class)
        if self._exempt_defaults:
            fields = {name: field for name, field in fields.items() if _is_required(field)}
        attribute_names = set(self._attributes.keys())
        model_attribute_names = set(fields.keys())
        missing_attributes = model_attribute_names - attribute_names
        if missing_attributes:
            raise ValueError(
                f"Attributes {missing_attributes} are not present in "
                f"{self._model_class.__name__}.\n"
                "Consider using strict=False"
            )

    def _optional_matchers(self, value: Any = None) -> list[Matcher]:
        if value is None:
            model_class = self._model_class
        else:
            model_class = type(value)
        if self._strict and self._exempt_defaults:
            fields = _model_fields(model_class)
            return [
                HasAttribute(name, _default_value(field))
                for name, field in fields.items()
                if not _is_required(field) and name not in self._attributes
            ]
        return []

    def _attribute_matchers(self, value: Any = None) -> list[Matcher]:
        return [
            HasAttribute(name, value) for name, value in self._attributes.items()
        ] + self._optional_matchers(value)

    def matches(self, value: Any) -> bool:
        return self._matches_instance(value) and all(
            attr_matcher == value for attr_matcher in self._attribute_matchers(value)
        )

    def _matches_instance(self, value: Any) -> bool:
        expected_model_version = _version(self._model_class)
        value_version = _version(type(value))
        return expected_model_version == value_version and self._is_instance_matcher == value

    def __repr__(self) -> str:
        return concat_reprs(
            f"To be a pydantic model instance of {repr(self._model_class.__name__)}",
            *self._attribute_matchers(),
        )



from abc import ABC, abstractmethod
from typing import Any


class BaseComparer(ABC):
    def __init__(self, actual: Any, expected: Any, *, fail_fast: bool = True):
        self._actual = actual
        self._expected = expected
        self._fail_fast = fail_fast

    @abstractmethod
    def compare(self) -> bool:
        pass

    def _base_compare(self, actual: Any, expected: Any) -> bool:
        # pylint: disable=import-outside-toplevel
        from pytest_matchers.asserts.comparer import Comparer

        return Comparer().compare(actual, expected, fail_fast=self._fail_fast)



from typing import Any

from pytest_matchers.asserts.comparers.dict import DictComparer
from pytest_matchers.asserts.comparers.list import ListComparer
from pytest_matchers.asserts.comparers.set import SetComparer
from pytest_matchers.matchers.base import Matcher


class Comparer:
    def compare(self, actual: Any, expected: Any, *, fail_fast: bool = True) -> bool:
        if isinstance(expected, Matcher) and not isinstance(actual, Matcher):
            return self.compare(expected, actual, fail_fast=fail_fast)  # Reverse the order
        return self._compare_by_type(actual, expected, fail_fast=fail_fast)

    def _compare_by_type(self, actual: Any, expected: Any, fail_fast: bool = True) -> bool:
        if isinstance(expected, (list, tuple)):
            return ListComparer(actual, expected, fail_fast=fail_fast).compare()
        if isinstance(expected, dict):
            return DictComparer(actual, expected, fail_fast=fail_fast).compare()
        if isinstance(expected, set):
            return SetComparer(actual, expected, fail_fast=fail_fast).compare()
        return actual == expected



import uuid
from typing import Any, Type

from pytest_matchers.matchers.base import Matcher
from pytest_matchers.matchers.is_instance import is_instance_matcher
from pytest_matchers.matchers.matcher_factory import matcher
from pytest_matchers.utils.matcher_utils import (
    as_matcher_or_none,
    matches_or_none,
)
from pytest_matchers.utils.repr_utils import concat_matcher_repr, concat_reprs


@matcher
class UUID(Matcher):
    def __init__(self, matching_type: Type = None, *, version: int | Matcher = None):
        super().__init__()
        self._is_instance_matcher = is_instance_matcher(matching_type)
        self._version_matcher = as_matcher_or_none(version)

    def matches(self, value: Any) -> bool:
        if matches_or_none(self._is_instance_matcher, value):
            try:
                uuid_value = uuid.UUID(str(value))
                return matches_or_none(self._version_matcher, uuid_value.version)
            except ValueError:
                pass
        return False

    def __repr__(self) -> str:
        base_repr = "To be a UUID"
        instance_matcher_repr = concat_matcher_repr(self._is_instance_matcher) or ""
        version_matcher_repr = (
            f"with version {concat_matcher_repr(self._version_matcher)}"
            if self._version_matcher
            else ""
        )
        return concat_reprs(f"{base_repr}", instance_matcher_repr, version_matcher_repr)



from pytest_matchers.asserts.comparers.base import BaseComparer


class ListComparer(BaseComparer):
    def compare(self) -> bool:
        same_length = len(self._actual) == len(self._expected)
        if not same_length and self._fail_fast:
            return False

        return self._items_compare() and same_length

    def _items_compare(self) -> bool:
        compared_items = (
            self._base_compare(actual_item, expected_item)
            for actual_item, expected_item in zip(self._actual, self._expected)
        )
        if not self._fail_fast:
            compared_items = list(compared_items)  # Force full generation
        return all(compared_items)



from typing import Any, Type

from pytest_matchers.matchers import IsInstance, Matcher
from pytest_matchers.matchers.is_instance import is_instance_matcher
from pytest_matchers.matchers.length import length_matcher
from pytest_matchers.matchers.matcher_factory import matcher
from pytest_matchers.utils.matcher_utils import (
    as_matcher_or_none,
    matches_or_none,
    partial_matches_or_none,
)
from pytest_matchers.utils.repr_utils import concat_reprs


@matcher
class List(Matcher):
    def __init__(
        self,
        match_type: Type | Matcher | None = None,
        *,
        length: int | None = None,
        min_length: int | None = None,
        max_length: int | None = None,
    ):
        super().__init__()
        self._type_matcher = (
            is_instance_matcher(match_type)
            if isinstance(match_type, type)
            else as_matcher_or_none(match_type)
        )
        self._length_matcher = length_matcher(length, min_length, max_length)

    def matches(self, value: Any) -> bool:
        return self._matches_type(value) and matches_or_none(self._length_matcher, value)

    def _matches_type(self, value: Any) -> bool:
        return IsInstance(list) == value and all(
            map(partial_matches_or_none(self._type_matcher), value)
        )

    def __repr__(self) -> str:
        return concat_reprs(
            "To be a list",
            self._type_matcher,
            self._length_matcher,
        )

    def concatenated_repr(self):
        return concat_reprs(
            "of lists",
            self._type_matcher,
            self._length_matcher,
        )



from typing import Any

from pytest_matchers.matchers.base import Matcher


class MatcherDetector:
    def __init__(self, value: Any):
        self._value = value

    def uses_matchers(self) -> bool:
        return isinstance(self._value, Matcher) or self._iterable_with_matchers()

    def _iterable_with_matchers(self) -> bool:
        if isinstance(self._value, dict):
            return any(MatcherDetector(value).uses_matchers() for value in self._value.values())
        if isinstance(self._value, (list, tuple, set)):
            return any(MatcherDetector(value).uses_matchers() for value in self._value)
        return False



from typing import Any

from pytest_matchers.matchers.base import Matcher
from pytest_matchers.matchers.matcher_factory import matcher
from pytest_matchers.utils.matcher_detector import MatcherDetector
from pytest_matchers.utils.repr_utils import as_matcher_repr


def _in_string_matcher(contained_value: Matcher, value: str):
    if contained_value == value:
        return True
    for i in range(len(value)):
        for j in range(i + 1, len(value) + 1):
            substring = value[i:j]
            if contained_value == substring:
                return True
    return False


@matcher
class Contains(Matcher):
    def __init__(self, *contained_values: Any):
        super().__init__()
        if len(contained_values) == 0:
            raise ValueError("At least one value must be provided")
        self._contained_values = contained_values

    def matches(self, value: Any) -> bool:
        try:
            return all(
                self._in(contained_value, value) for contained_value in self._contained_values
            )
        except TypeError:
            return False

    @staticmethod
    def _in(contained_value: Any, value: Any) -> bool:
        if MatcherDetector(contained_value).uses_matchers() and isinstance(value, str):
            return _in_string_matcher(contained_value, value)
        return contained_value in value

    def __repr__(self) -> str:
        return f"To contain {', '.join(map(repr, self._contained_values))}"

    def concatenated_repr(self) -> str:
        return (
            "containing something expected "
            f"{', and something '.join(map(as_matcher_repr, self._contained_values))}"
        )


def contains_matcher(contains: str | None) -> Contains | None:
    if contains is None:
        return None
    return Contains(contains)



from typing import Any

from pytest_matchers.asserts.comparer import Comparer


def _match(actual: Any, expected: Any) -> bool:
    return Comparer().compare(actual, expected)


def assert_match(actual: Any, expected: Any) -> None:
    try:
        assert _match(actual, expected)
    except AssertionError:
        assert actual == expected  # Return the original assertion error


def _not_match(actual: Any, expected: Any) -> bool:
    return not _match(actual, expected)


def assert_not_match(actual: Any, expected: Any) -> None:
    try:
        assert _not_match(actual, expected)
    except AssertionError:
        assert actual != expected  # Return the original assertion error



from typing import Any, Callable, Type

from pytest_matchers.matchers import (
    JSON,
    UUID,
    AfterApplying,
    And,
    Anything,
    Between,
    Case,
    Contains,
    Datetime,
    DatetimeString,
    Dict,
    DifferentValue,
    ExceptionMatcher,
    HasAttribute,
    If,
    IsInstance,
    List,
    Matcher,
    Number,
    Or,
    SameValue,
    StrictDict,
    String,
    Timestamp,
)


def anything() -> Anything:
    return Anything()


def is_instance(match_type: Type) -> IsInstance:
    return IsInstance(match_type)


def is_list(match_type: Type | Matcher | None = None, **kwargs) -> List | IsInstance:
    if match_type is None and not kwargs:
        return is_instance(list)
    return List(match_type, **kwargs)


def is_string(**kwargs) -> String:
    return String(**kwargs)


def not_empty_string(**kwargs) -> String:
    min_length = 1
    if "min_length" in kwargs:
        min_length = kwargs.pop("min_length")

    return is_string(min_length=min_length, **kwargs)


def is_number(match_type: Type = None, **kwargs) -> Number:
    return Number(match_type, **kwargs)


def is_int(**kwargs) -> Number:
    return is_number(int, **kwargs)


def is_float(**kwargs) -> Number:
    return is_number(float, **kwargs)


def one_of(*values: Matcher | Any) -> Or:
    return Or(*values)


def has_attribute(attribute_name: str, expected_value: Any = None) -> HasAttribute:
    return HasAttribute(attribute_name, expected_value)


def between(
    min_value: Any,
    max_value: Any,
    *,
    inclusive: bool = None,
    min_inclusive: bool = None,
    max_inclusive: bool = None,
) -> Between:
    return Between(min_value, max_value, inclusive, min_inclusive, max_inclusive)


def contains(*values: Any) -> And:
    return And(*[Contains(value) for value in values])


def is_datetime(
    min_value: Any = None,
    max_value: Any = None,
    year: int = None,
    month: int = None,
    day: int = None,
    hour: int = None,
    minute: int = None,
    second: int = None,
) -> Datetime:
    return Datetime(min_value, max_value, year, month, day, hour, minute, second)


def is_datetime_string(
    expected_format: str,
    *,
    min_value: Any = None,
    max_value: Any = None,
) -> DatetimeString:
    return DatetimeString(expected_format, min_value=min_value, max_value=max_value)


def is_timestamp(min_value: float | Any = None, max_value: float | Any = None) -> Timestamp:
    return Timestamp(min_value=min_value, max_value=max_value)


def is_iso_8601_date(**kwargs) -> DatetimeString:
    return is_datetime_string("%Y-%m-%d", **kwargs)


def is_iso_8601_datetime(**kwargs) -> DatetimeString:
    return is_datetime_string("%Y-%m-%dT%H:%M:%S", **kwargs)


def is_iso_8601_time(**kwargs) -> DatetimeString:
    return is_datetime_string("%H:%M:%S", **kwargs)


def same_value() -> SameValue:
    return SameValue()


def different_value() -> DifferentValue:
    return DifferentValue()


def if_true(
    condition: Callable | bool | Any,
    then: Matcher | Any = None,
    or_else: Matcher | Any = None,
) -> If:
    return If(condition, then, or_else)


def if_false(
    condition: Callable | bool | Any,
    then: Matcher | Any = None,
    or_else: Matcher | Any = None,
) -> If:
    return if_true(condition, or_else, then)  # pylint: disable=arguments-out-of-order


def case(
    case_value: Any,
    expectations: dict[Any, Matcher | Any],
    default_expectation: Matcher | Any | None = None,
) -> Case:
    return Case(case_value, expectations, default_expectation)


def is_dict(matching: dict = None, *, exclude: list[Any] = None) -> Dict:
    return Dict(matching, exclude=exclude)


def is_strict_dict(
    matching: dict,
    *,
    extra_condition: bool = None,
    when_true: dict = None,
    when_false: dict = None,
) -> StrictDict:
    return StrictDict(
        matching,
        extra_condition=extra_condition,
        when_true=when_true,
        when_false=when_false,
    )


def is_json(matching: dict = None, *, exclude: list[Any] = None) -> JSON:
    return JSON(matching, exclude=exclude)


def is_uuid(matching_type: Type = None, *, version: int | Matcher = None) -> UUID:
    return UUID(matching_type, version=version)


def is_exception(
    exception_type: Type[Exception] = Exception,
    message: str | Matcher | None = None,
    *,
    match_subclass: bool = False,
) -> ExceptionMatcher:
    return ExceptionMatcher(exception_type, message, match_subclass)


def after_applying_is(function: Callable, expected_value: Any | Matcher) -> AfterApplying:
    return AfterApplying(function, expected_value)



from typing import Generator

from pytest_matchers.asserts.comparers.base import BaseComparer


class DictComparer(BaseComparer):
    def compare(self) -> bool:
        same_length = len(self._actual) == len(self._expected)
        if not same_length and self._fail_fast:
            return False

        return self._items_compare() and same_length

    def _items_compare(self) -> bool:
        compared_items = self._compared_items()
        if not self._fail_fast:
            compared_items = list(compared_items)
        return all(compared_items)

    def _compared_items(self) -> Generator:
        for key in self._expected:
            if key in self._actual:
                yield self._base_compare(self._actual[key], self._expected[key])
            else:
                yield False



from collections.abc import Sized
from typing import Any

from pytest_matchers.matchers.base import Matcher
from pytest_matchers.matchers.matcher_factory import matcher
from pytest_matchers.utils.matcher_detector import MatcherDetector
from pytest_matchers.utils.repr_utils import as_matcher_repr


def _starts_with_matcher(prefix: Matcher, value: Any):
    if prefix == value:  # With some luck the matcher matches with the whole value
        return True
    try:
        for index in range(1, len(value)):
            if prefix == value[0:index]:
                return True
    except TypeError:
        pass
    return False


@matcher
class StartsWith(Matcher):
    def __init__(self, prefix: Sized | str | Matcher):
        super().__init__()
        self._prefix = prefix

    def matches(self, value: Any) -> bool:
        if MatcherDetector(self._prefix).uses_matchers():
            return _starts_with_matcher(self._prefix, value)

        if isinstance(value, str):
            try:
                return value.startswith(self._prefix)
            except TypeError:
                return False
        if isinstance(value, list):
            return value[: len(self._prefix)] == self._prefix

        return False

    def __repr__(self) -> str:
        return f"To start with {repr(self._prefix)}"

    def concatenated_repr(self) -> str:
        return f"with start expected {as_matcher_repr(self._prefix)}"


def starts_with_matcher(starts_with: str | None) -> StartsWith | None:
    if starts_with is None:
        return None
    return StartsWith(starts_with)
